@@ -216,12 +216,12 @@ highlight_at_position(Column, Severity, Length) ->
 
 highlight_below_line(Line, Severity) ->
   % Don't highlight leading whitespaces in line
-  {_, SpacesMatched} = trim_line(Line, 0),
+  {Rest, SpacesMatched} = trim_line(Line, 0),
 
-  Length = string:length(Line),
+  Length = string:length(Rest),
   Highlight = case Severity of
-    warning -> highlight(lists:duplicate(Length - SpacesMatched, $~), warning);
-    error -> highlight(lists:duplicate(Length - SpacesMatched, $^), error)
+    warning -> highlight(lists:duplicate(Length, $~), warning);
+    error -> highlight(lists:duplicate(Length, $^), error)
   end,
 
   [n_spaces(SpacesMatched), Highlight].
@@ -348,7 +348,7 @@ parse_error(Location, File, <<"syntax error before: ">>, Keyword, Input)
 
 %% Produce a human-readable message for errors before a sigil
 parse_error(Location, File, <<"syntax error before: ">>, <<"{sigil,", _Rest/binary>> = Full, Input) ->
-  {sigil, _, Atom, [Content | _], _, _, _} = parse_erl_term(Full),
+  {ok, {sigil, _, Atom, [Content | _], _, _, _}} = parse_erl_term(Full),
   Content2 = case is_binary(Content) of
     true -> Content;
     false -> <<>>
@@ -369,7 +369,7 @@ parse_error(Location, File, <<"syntax error before: ">>, <<"{sigil,", _Rest/bina
 %% Binaries (and interpolation) are wrapped in [<<...>>]
 parse_error(Location, File, Error, <<"[", _/binary>> = Full, Input) when is_binary(Error) ->
   Term = case parse_erl_term(Full) of
-    [H | _] when is_binary(H) -> <<$", H/binary, $">>;
+    {ok, [H | _]} when is_binary(H) -> <<$", H/binary, $">>;
     _ -> <<$">>
   end,
   raise_snippet(Location, File, Input, 'Elixir.SyntaxError', <<Error/binary, Term/binary>>);
@@ -399,9 +399,14 @@ parse_error(Location, File, Error, Token, Input) when is_binary(Error), is_binar
   end.
 
 parse_erl_term(Term) ->
-  {ok, Tokens, _} = erl_scan:string(binary_to_list(Term)),
-  {ok, Parsed} = erl_parse:parse_term(Tokens ++ [{dot, 1}]),
-  Parsed.
+  case erl_scan:string(binary_to_list(Term)) of
+    {ok, Tokens, _} ->
+      case erl_parse:parse_term(Tokens ++ [{dot, 1}]) of
+        {ok, Parsed} -> {ok, Parsed};
+        _ -> error
+      end;
+    _ -> error
+  end.
 
 raise_reserved(Location, File, Input, Keyword) ->
   raise_snippet(Location, File, Input, 'Elixir.SyntaxError',
