@@ -188,6 +188,71 @@ defmodule Kernel.ExpansionTest do
       assert output == quote(do: _ = 1)
       assert Macro.Env.vars(env) == []
     end
+
+    test "errors on directly recursive definitions" do
+      assert_compile_error(
+        ~r"""
+        recursive variable definition in patterns:
+
+        \{x = \{:ok, x\}\}
+
+        the variable "x" \(context Kernel.ExpansionTest\) is defined in function of itself
+        """,
+        fn -> expand(quote(do: {x = {:ok, x}} = :ok)) end
+      )
+
+      assert_compile_error(
+        ~r"""
+        recursive variable definition in patterns:
+
+        \{\{x, y\} = \{y, x\}\}
+
+        the variable "x" \(context Kernel.ExpansionTest\) is defined in function of itself
+        """,
+        fn -> expand(quote(do: {{x, y} = {y, x}} = :ok)) end
+      )
+
+      assert_compile_error(
+        ~r"""
+        recursive variable definition in patterns:
+
+        \{\{:x, y\} = \{x, :y\}, x = y\}
+
+        the variable "x" \(context Kernel.ExpansionTest\) is defined recursively in function of "y" \(context Kernel.ExpansionTest\)
+        """,
+        fn -> expand(quote(do: {{:x, y} = {x, :y}, x = y} = :ok)) end
+      )
+
+      assert_compile_error(
+        ~r"""
+        recursive variable definition in patterns:
+
+        \{x = y, y = z, z = x\}
+
+        the following variables form a cycle: "x" \(context Kernel.ExpansionTest\), "y" \(context Kernel.ExpansionTest\), "z" \(context Kernel.ExpansionTest\)
+        """,
+        fn -> expand(quote(do: {x = y, y = z, z = x} = :ok)) end
+      )
+    end
+
+    test "complex recursive variable definitions" do
+      assert expand(
+               quote do:
+                       {%{type: type, client_id: client_id} = message,
+                        %{type: type, client_id: client_id} = state} = :ok
+             )
+
+      assert_compile_error(
+        ~r"recursive variable definition in patterns",
+        fn ->
+          expand(
+            quote do:
+                    {%{type: type, client_id: client_id} = message,
+                     %{type: type, client_id: client_id} = state, client_id = type} = :ok
+          )
+        end
+      )
+    end
   end
 
   describe "environment macros" do
@@ -654,12 +719,6 @@ defmodule Kernel.ExpansionTest do
     test "expands base and args" do
       assert expand(quote(do: a.(b))) == quote(do: a().(b()))
     end
-
-    test "raises on atom base" do
-      assert_compile_error(~r"invalid function call :foo.()", fn ->
-        expand(quote(do: :foo.(a)))
-      end)
-    end
   end
 
   describe "remotes" do
@@ -954,6 +1013,11 @@ defmodule Kernel.ExpansionTest do
         ~r"when using :reduce with comprehensions, the do block must be written using acc -> expr clauses",
         fn -> expand(quote(do: for(x <- 1..3, reduce: %{}, do: (acc, x -> x)))) end
       )
+
+      assert_compile_error(
+        ~r"when using :reduce with comprehensions, the do block must be written using acc -> expr clauses",
+        fn -> expand(quote(do: for(x <- 1..3, reduce: %{}, do: (acc, x when 1 == 1 -> x)))) end
+      )
     end
 
     test "raise error for unknown options" do
@@ -1905,6 +1969,32 @@ defmodule Kernel.ExpansionTest do
       assert expand(before_expansion) == after_expansion
     end
 
+    test "expands catch with when" do
+      before_expansion =
+        quote do
+          try do
+            x
+          catch
+            x when x -> z = :erlang.-(x)
+          end
+
+          z
+        end
+
+      after_expansion =
+        quote do
+          try do
+            x()
+          catch
+            :throw, x when x -> z = :erlang.-(x)
+          end
+
+          z()
+        end
+
+      assert expand(before_expansion) == after_expansion
+    end
+
     test "expands after" do
       before_expansion =
         quote do
@@ -2136,6 +2226,19 @@ defmodule Kernel.ExpansionTest do
 
         expand(code)
       end)
+
+      assert_compile_error(message, fn ->
+        code =
+          quote do
+            try do
+              x
+            catch
+              _, _, _ when 1 == 1 -> :ok
+            end
+          end
+
+        expand(code)
+      end)
     end
 
     test "expects clauses for rescue, else, catch" do
@@ -2274,81 +2377,9 @@ defmodule Kernel.ExpansionTest do
                quote(do: <<foo::integer>> = baz = <<bar()::integer>>)
                |> clean_bit_modifiers()
 
-      assert expand(quote(do: <<foo>> = {<<baz>>} = bar())) |> clean_meta([:alignment]) ==
-               quote(do: <<foo::integer>> = {<<baz::integer>>} = bar())
+      assert expand(quote(do: <<foo>> = <<bar>> = baz)) |> clean_meta([:alignment]) ==
+               quote(do: <<foo::integer>> = <<bar::integer>> = baz())
                |> clean_bit_modifiers()
-
-      message = ~r"binary patterns cannot be matched in parallel using \"=\""
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: <<foo>> = <<baz>> = bar()))
-      end)
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: <<foo>> = qux = <<baz>> = bar()))
-      end)
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: {<<foo>>} = {qux} = {<<baz>>} = bar()))
-      end)
-
-      assert expand(quote(do: {:foo, <<foo>>} = {<<baz>>, :baz} = bar()))
-
-      # two-element tuples are special cased
-      assert_compile_error(message, fn ->
-        expand(quote(do: {:foo, <<foo>>} = {:foo, <<baz>>} = bar()))
-      end)
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: %{foo: <<foo>>} = %{baz: <<qux>>, foo: <<baz>>} = bar()))
-      end)
-
-      assert expand(quote(do: %{foo: <<foo>>} = %{baz: <<baz>>} = bar()))
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: %_{foo: <<foo>>} = %_{foo: <<baz>>} = bar()))
-      end)
-
-      assert expand(quote(do: %_{foo: <<foo>>} = %_{baz: <<baz>>} = bar()))
-
-      assert_compile_error(message, fn ->
-        expand(quote(do: %_{foo: <<foo>>} = %{foo: <<baz>>} = bar()))
-      end)
-
-      assert expand(quote(do: %_{foo: <<foo>>} = %{baz: <<baz>>} = bar()))
-
-      assert_compile_error(message, fn ->
-        code =
-          quote do
-            case bar() do
-              <<foo>> = <<baz>> -> nil
-            end
-          end
-
-        expand(code)
-      end)
-
-      assert_compile_error(message, fn ->
-        code =
-          quote do
-            case bar() do
-              <<foo>> = qux = <<baz>> -> nil
-            end
-          end
-
-        expand(code)
-      end)
-
-      assert_compile_error(message, fn ->
-        code =
-          quote do
-            case bar() do
-              [<<foo>>] = [<<baz>>] -> nil
-            end
-          end
-
-        expand(code)
-      end)
     end
 
     test "nested match" do
@@ -2800,16 +2831,6 @@ defmodule Kernel.ExpansionTest do
       end)
     end
 
-    test "raises for invalid literals" do
-      assert_compile_error(~r"invalid literal :foo in <<>>", fn ->
-        expand(quote(do: <<:foo>>))
-      end)
-
-      assert_compile_error(~r"invalid literal \[\] in <<>>", fn ->
-        expand(quote(do: <<[]::size(8)>>))
-      end)
-    end
-
     test "raises on binary fields with size in matches" do
       assert expand(quote(do: <<x::binary-size(3), y::binary>> = "foobar"))
 
@@ -2865,11 +2886,11 @@ defmodule Kernel.ExpansionTest do
 
   test "handles invalid expressions" do
     assert_compile_error(~r"invalid quoted expression: {1, 2, 3}", fn ->
-      expand_env(quote(do: unquote({1, 2, 3})), __ENV__)
+      expand_env({1, 2, 3}, __ENV__)
     end)
 
     assert_compile_error(~r"invalid quoted expression: #Function\<", fn ->
-      expand(quote(do: unquote({:sample, fn -> nil end})))
+      expand({:sample, fn -> nil end})
     end)
 
     assert_compile_error(~r"invalid pattern in match", fn ->
@@ -2913,43 +2934,6 @@ defmodule Kernel.ExpansionTest do
     assert_compile_error(~r"misplaced operator ->", fn ->
       expand(quote(do: (foo -> bar)))
     end)
-
-    message = ~r/"wrong_fun" cannot handle clauses with the ->/
-
-    assert_compile_error(message, fn ->
-      code =
-        quote do
-          wrong_fun do
-            _ -> :ok
-          end
-        end
-
-      expand(code)
-    end)
-
-    assert_compile_error(message, fn ->
-      code =
-        quote do
-          wrong_fun do
-            foo -> bar
-          after
-            :ok
-          end
-        end
-
-      expand(code)
-    end)
-
-    assert_compile_error(~r/"length" cannot handle clauses with the ->/, fn ->
-      code =
-        quote do
-          length do
-            _ -> :ok
-          end
-        end
-
-      expand(code)
-    end)
   end
 
   ## Helpers
