@@ -422,8 +422,6 @@ defmodule Kernel.ParallelCompilerTest do
     end
 
     test "supports warnings as errors" do
-      warnings_as_errors = Code.get_compiler_option(:warnings_as_errors)
-
       [fixture] =
         write_tmp(
           "warnings_as_errors",
@@ -438,20 +436,19 @@ defmodule Kernel.ParallelCompilerTest do
       output = tmp_path("not_to_be_used")
 
       try do
-        Code.compiler_options(warnings_as_errors: true)
-
         msg =
           capture_io(:stderr, fn ->
             assert {:error, [error], []} =
-                     Kernel.ParallelCompiler.compile_to_path([fixture], output)
+                     Kernel.ParallelCompiler.compile_to_path([fixture], output,
+                       warnings_as_errors: true
+                     )
 
-            assert {^fixture, 3, "this clause " <> _} = error
+            assert {^fixture, {3, 7}, "this clause " <> _} = error
           end)
 
         assert msg =~
                  "Compilation failed due to warnings while using the --warnings-as-errors option\n"
       after
-        Code.compiler_options(warnings_as_errors: warnings_as_errors)
         purge([WarningsSample])
       end
 
@@ -465,18 +462,18 @@ defmodule Kernel.ParallelCompilerTest do
         write_tmp(
           "error_line",
           a: """
-          defmodule A do
+          defmodule PCA do
             def fun(arg), do: arg / 2
           end
           """,
           b: """
-          defmodule B do
+          defmodule PCB do
             def fun(arg) do
-              A.fun(arg)
+              PCA.fun(arg)
               :ok
             end
           end
-          B.fun(:not_a_number)
+          PCB.fun(:not_a_number)
           """
         )
 
@@ -592,8 +589,6 @@ defmodule Kernel.ParallelCompilerTest do
     end
 
     test "supports warnings as errors" do
-      warnings_as_errors = Code.get_compiler_option(:warnings_as_errors)
-
       [fixture] =
         write_tmp(
           "warnings_as_errors",
@@ -606,19 +601,17 @@ defmodule Kernel.ParallelCompilerTest do
         )
 
       try do
-        Code.compiler_options(warnings_as_errors: true)
-
         msg =
           capture_io(:stderr, fn ->
-            assert {:error, [error], []} = Kernel.ParallelCompiler.require([fixture])
+            assert {:error, [error], []} =
+                     Kernel.ParallelCompiler.require([fixture], warnings_as_errors: true)
 
-            assert {^fixture, 3, "this clause " <> _} = error
+            assert {^fixture, {3, 7}, "this clause " <> _} = error
           end)
 
         assert msg =~
                  "Compilation failed due to warnings while using the --warnings-as-errors option\n"
       after
-        Code.compiler_options(warnings_as_errors: warnings_as_errors)
         purge([WarningsSample])
       end
     end
