@@ -38,23 +38,166 @@ defmodule Module.Types.IntegrationTest do
 
       modules = compile(files)
 
-      assert read_chunk(modules[A]).exports == [
-               {{:c, 0}, %{deprecated_reason: nil, kind: :def}},
-               {{:d, 0}, %{deprecated_reason: nil, kind: :defmacro}},
-               {{:e, 0}, %{deprecated_reason: "oops", kind: :def}}
-             ]
+      assert [
+               {{:c, 0}, %{}},
+               {{:e, 0}, %{deprecated: "oops", sig: {:infer, _}}}
+             ] = read_chunk(modules[A]).exports
 
       assert read_chunk(modules[B]).exports == [
-               {{:behaviour_info, 1}, %{deprecated_reason: nil, kind: :def}}
+               {{:behaviour_info, 1}, %{sig: :none}}
              ]
 
       assert read_chunk(modules[C]).exports == [
-               {{:behaviour_info, 1}, %{deprecated_reason: nil, kind: :def}}
+               {{:behaviour_info, 1}, %{sig: :none}}
              ]
     end
   end
 
-  describe "undefined" do
+  describe "type checking" do
+    test "inferred remote calls" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def fun(:ok), do: :doki
+          def fun(:error), do: :bad
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          def badarg do
+            A.fun(:unknown)
+          end
+
+          def badmatch do
+            :doki = A.fun(:error)
+          end
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible types given to A.fun/1:
+
+                A.fun(:unknown)
+        """,
+        """
+            but expected one of:
+
+                #1
+                dynamic(:ok)
+
+                #2
+                dynamic(:error)
+        """,
+        """
+            warning: the following pattern will never match:
+
+                :doki = A.fun(:error)
+
+            because the right-hand side has type:
+
+                dynamic(:bad)
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "mismatched locals" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def error(), do: private(raise "oops")
+          def public(x), do: private(List.to_tuple(x))
+          defp private(:ok), do: nil
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible types given to private/1:
+
+                private(raise RuntimeError.exception("oops"))
+
+        """,
+        "the 1st argument is empty (often represented as none())",
+        """
+            typing violation found at:
+            │
+          2 │   def error(), do: private(raise "oops")
+            │                    ~
+            │
+            └─ a.ex:2:20: A.error/0
+        """,
+        """
+            warning: incompatible types given to private/1:
+
+                private(List.to_tuple(x))
+        """,
+        """
+            typing violation found at:
+            │
+          3 │   def public(x), do: private(List.to_tuple(x))
+            │                      ~
+            │
+            └─ a.ex:3:22: A.public/1
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    test "unused private clauses" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def public(x) do
+            private(List.to_tuple(x))
+          end
+
+          defp private(nil), do: nil
+          defp private("foo"), do: "foo"
+          defp private({:ok, ok}), do: ok
+          defp private({:error, error}), do: error
+          defp private("bar"), do: "bar"
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: this clause of defp private/1 is never used
+            │
+          6 │   defp private(nil), do: nil
+            │        ~
+            │
+            └─ a.ex:6:8: A.private/1
+        """,
+        """
+            warning: this clause of defp private/1 is never used
+            │
+          7 │   defp private("foo"), do: "foo"
+            │        ~
+            │
+            └─ a.ex:7:8: A.private/1
+        """,
+        """
+            warning: this clause of defp private/1 is never used
+            │
+         10 │   defp private("bar"), do: "bar"
+            │        ~
+            │
+            └─ a.ex:10:8: A.private/1
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+  end
+
+  describe "undefined warnings" do
     test "handles Erlang modules" do
       files = %{
         "a.ex" => """
@@ -296,42 +439,6 @@ defmodule Module.Types.IntegrationTest do
       assert_warnings(files, warnings)
     end
 
-    test "protocols are checked, ignoring missing built-in impls" do
-      files = %{
-        "a.ex" => """
-        defprotocol AProtocol do
-          def func(arg)
-        end
-
-        defmodule AImplementation do
-          defimpl AProtocol do
-            def func(_), do: B.no_func()
-          end
-        end
-        """
-      }
-
-      warnings = [
-        "B.no_func/0 is undefined (module B is not available or is yet to be defined)",
-        "a.ex:7:24: AProtocol.AImplementation.func/1"
-      ]
-
-      assert_warnings(files, warnings)
-    end
-
-    test "handles Erlang ops" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          def a(a, b), do: a and b
-          def b(a, b), do: a or b
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-
     test "hints exclude deprecated functions" do
       files = %{
         "a.ex" => """
@@ -387,7 +494,7 @@ defmodule Module.Types.IntegrationTest do
       }
 
       warnings = [
-        "you must require B before invoking the macro B.b/0",
+        "Be sure to require B if you intend to invoke this macro",
         "ab.ex:2:18: A.a/0"
       ]
 
@@ -604,6 +711,23 @@ defmodule Module.Types.IntegrationTest do
   end
 
   describe "regressions" do
+    test "does not emit false positives from defguard" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          defguard is_non_nil_arity_function(fun, arity)
+                   when arity != nil and is_function(fun, arity)
+
+          def check(fun, args) do
+            is_non_nil_arity_function(fun, length(args))
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+
     test "do not parse binary segments as variables" do
       files = %{
         "a.ex" => """
@@ -824,8 +948,8 @@ defmodule Module.Types.IntegrationTest do
 
     Map.new(modules, fn module ->
       {^module, binary, _filename} = :code.get_object_code(module)
-      :code.purge(module)
       :code.delete(module)
+      :code.purge(module)
       {module, binary}
     end)
   end
