@@ -121,6 +121,12 @@ defmodule ExUnit.Assertions do
 
   Even though the match works, `assert` still expects a truth
   value. In such cases, simply use `==/2` or `match?/2`.
+
+  If you need more complex pattern matching using guards, you
+  need to use `match?/2`:
+
+      assert match?([%{id: id} | _] when is_integer(id), records)
+
   """
   defmacro assert({:=, meta, [left, right]} = assertion) do
     code = escape_quoted(:assert, meta, assertion)
@@ -223,7 +229,7 @@ defmodule ExUnit.Assertions do
 
   defmacro refute(assertion) do
     if translated = translate_assertion(:refute, assertion, __CALLER__) do
-      {:!, [], [translated]}
+      {:!, [generated: true], [translated]}
     else
       {args, value} = extract_args(assertion, __CALLER__)
 
@@ -357,6 +363,23 @@ defmodule ExUnit.Assertions do
   end
 
   @doc false
+  def __match__({:when, _, _} = left, right, _, _, _) do
+    suggestion =
+      quote do
+        assert match?(unquote(left), unquote(right))
+      end
+
+    raise ArgumentError, """
+    invalid pattern in assert/1:
+
+    #{Macro.to_string(left) |> Inspect.Error.pad(2)}
+
+    To assert with guards, use match?/2:
+
+    #{Macro.to_string(suggestion) |> Inspect.Error.pad(2)}
+    """
+  end
+
   def __match__(left, right, code, check, caller) do
     left = __expand_pattern__(left, caller)
     vars = collect_vars_from_pattern(left)
@@ -410,7 +433,7 @@ defmodule ExUnit.Assertions do
   end
 
   def assert(value, opts) when is_list(opts) do
-    unless value, do: raise(ExUnit.AssertionError, opts)
+    if !value, do: raise(ExUnit.AssertionError, opts)
     true
   end
 
@@ -509,8 +532,12 @@ defmodule ExUnit.Assertions do
         end
       end
 
-    failure_message =
-      failure_message ||
+    on_timeout =
+      if failure_message do
+        quote do
+          flunk(unquote(failure_message))
+        end
+      else
         quote do
           ExUnit.Assertions.__timeout__(
             unquote(Macro.escape(expanded_pattern)),
@@ -520,6 +547,7 @@ defmodule ExUnit.Assertions do
             timeout
           )
         end
+      end
 
     quote do
       timeout = ExUnit.Assertions.__timeout__(unquote(timeout), :assert_receive_timeout)
@@ -528,7 +556,7 @@ defmodule ExUnit.Assertions do
         receive do
           unquote(pattern) -> {received, unquote(mark_as_generated(vars))}
         after
-          timeout -> flunk(unquote(failure_message))
+          timeout -> unquote(on_timeout)
         end
 
       received
@@ -780,7 +808,7 @@ defmodule ExUnit.Assertions do
         "expected:\n  #{inspect(message)}\n" <>
         "actual:\n" <> "  #{inspect(Exception.message(error))}"
 
-    unless match?, do: flunk(message)
+    if not match?, do: flunk(message)
 
     error
   end
