@@ -431,7 +431,7 @@ defmodule ExUnitTest do
   test "raises friendly error for duplicate test names" do
     message = ~S("test duplicate" is already defined in ExUnitTest.TestWithSameNames)
 
-    assert_raise ExUnit.DuplicateTestError, message, fn ->
+    assert_raise ArgumentError, message, fn ->
       defmodule TestWithSameNames do
         use ExUnit.Case
 
@@ -683,6 +683,38 @@ defmodule ExUnitTest do
     assert output =~ "\n6 tests, 0 failures, 1 excluded, 4 invalid, 1 skipped\n"
   end
 
+  test "parameterized tests" do
+    Process.register(self(), :parameterized_tests)
+
+    defmodule ParameterizedTests do
+      use ExUnit.Case, async: true, parameterize: [%{value: true}, %{value: false}]
+
+      @tag :tmp_dir
+      test "hello world", %{value: value, tmp_dir: tmp_dir} do
+        send(:parameterized_tests, {:tmp_dir, tmp_dir})
+        assert value
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true)
+
+    output = capture_io(fn -> ExUnit.run() end)
+
+    assert output =~ ~r"""
+           ExUnitTest.ParameterizedTests \[.*test/ex_unit_test.exs\]
+           Parameters: %\{value: false\}
+           """
+
+    assert output =~ """
+             1) test hello world (ExUnitTest.ParameterizedTests)
+                Parameters: %{value: false}
+           """
+
+    # Check that the temporary paths are different
+    assert_receive {:tmp_dir, tmp_dir1}
+    assert_receive {:tmp_dir, tmp_dir2} when tmp_dir1 != tmp_dir2
+  end
+
   describe "after_suite/1" do
     test "executes all callbacks set in reverse order" do
       Process.register(self(), :after_suite_test_process)
@@ -1030,10 +1062,137 @@ defmodule ExUnitTest do
     assert output =~ "2 tests, 0 failures, 2 excluded\n"
   end
 
+  test "tests are run in compile order (FIFO)" do
+    defmodule FirstTestFIFO do
+      use ExUnit.Case
+
+      test "first test" do
+        assert true
+      end
+    end
+
+    defmodule SecondTestFIFO do
+      use ExUnit.Case
+
+      test "second test" do
+        assert true
+      end
+    end
+
+    defmodule ThirdTestFIFO do
+      use ExUnit.Case
+
+      test "third test" do
+        assert true
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true)
+
+    output =
+      capture_io(fn ->
+        assert ExUnit.run() == %{total: 3, failures: 0, excluded: 0, skipped: 0}
+      end)
+
+    [_, first, second, third | _] = String.split(output, "\n\n")
+
+    assert first =~ "FirstTestFIFO"
+    assert second =~ "SecondTestFIFO"
+    assert third =~ "ThirdTestFIFO"
+  end
+
+  test "groups are run in compile order (FIFO)" do
+    defmodule RedOneFIFO do
+      use ExUnit.Case, async: true, group: :red
+
+      test "red one test" do
+        Process.sleep(5)
+        assert true
+      end
+    end
+
+    defmodule BlueOneFIFO do
+      use ExUnit.Case, async: true, group: :blue
+
+      test "blue one test" do
+        Process.sleep(5)
+        assert true
+      end
+    end
+
+    defmodule RedTwoFIFO do
+      use ExUnit.Case, async: true, group: :red
+
+      test "red two test" do
+        assert true
+      end
+    end
+
+    defmodule BlueTwoFIFO do
+      use ExUnit.Case, async: true, group: :blue
+
+      test "blue two test" do
+        assert true
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true, max_cases: 2)
+
+    output =
+      capture_io(fn ->
+        assert ExUnit.run() == %{total: 4, failures: 0, excluded: 0, skipped: 0}
+      end)
+
+    [_, first, second, third, fourth | _] = String.split(output, "\n\n")
+
+    assert first =~ "RedOneFIFO"
+    assert second =~ "RedTwoFIFO"
+    assert third =~ "BlueOneFIFO"
+    assert fourth =~ "BlueTwoFIFO"
+  end
+
+  test "filters async tests" do
+    defmodule FirstTestAsyncTrue do
+      use ExUnit.Case, async: true
+
+      test "first test" do
+        assert true
+      end
+    end
+
+    defmodule SecondTestAsyncTrue do
+      use ExUnit.Case, async: true
+
+      test "second test" do
+        assert true
+      end
+    end
+
+    defmodule FirstTestAsyncFalse do
+      use ExUnit.Case, async: false
+
+      test "first test" do
+        assert true
+      end
+    end
+
+    assert {%{failures: 0, skipped: 0, total: 3, excluded: 1}, _} =
+             run_with_filter([include: [async: true], exclude: [:test]], [])
+
+    assert {%{failures: 0, skipped: 0, total: 3, excluded: 2}, _} =
+             run_with_filter(
+               [include: [async: false], exclude: [:test]],
+               [FirstTestAsyncTrue, SecondTestAsyncTrue, FirstTestAsyncFalse]
+             )
+  end
+
   ##  Helpers
 
   defp run_with_filter(filters, cases) do
-    Enum.each(cases, &ExUnit.Server.add_sync_module/1)
+    Enum.each(cases, fn mod ->
+      ExUnit.Server.add_module(mod, mod.__ex_unit__(:config))
+    end)
+
     ExUnit.Server.modules_loaded(false)
 
     opts =
@@ -1041,8 +1200,7 @@ defmodule ExUnitTest do
       |> Keyword.merge(filters)
       |> Keyword.merge(colors: [enabled: false])
 
-    output = capture_io(fn -> Process.put(:capture_result, ExUnit.Runner.run(opts, nil)) end)
-    {Process.get(:capture_result) |> elem(0), output}
+    with_io(fn -> ExUnit.Runner.run(opts, nil) |> elem(0) end)
   end
 
   defp next_message_in_mailbox() do
