@@ -1,7 +1,7 @@
 defmodule Mix.Compilers.Elixir do
   @moduledoc false
 
-  @manifest_vsn 25
+  @manifest_vsn 26
   @checkpoint_vsn 2
 
   import Record
@@ -202,6 +202,13 @@ defmodule Mix.Compilers.Elixir do
 
           put_compile_env(sources)
           all_warnings = previous_warnings ++ runtime_warnings ++ compile_warnings
+
+          lazy_modules_diff = fn ->
+            modules_diff(modules, removed_modules, all_modules, timestamp)
+          end
+
+          Mix.Task.Compiler.notify_modules_compiled(lazy_modules_diff)
+
           unless_previous_warnings_as_errors(previous_warnings, opts, {:ok, all_warnings})
 
         {:error, errors, %{runtime_warnings: r_warnings, compile_warnings: c_warnings}, state} ->
@@ -247,6 +254,7 @@ defmodule Mix.Compilers.Elixir do
 
       all_warnings = Keyword.get(opts, :all_warnings, true)
       previous_warnings = previous_warnings(sources, all_warnings)
+
       unless_previous_warnings_as_errors(previous_warnings, opts, {status, previous_warnings})
     end
   end
@@ -316,6 +324,14 @@ defmodule Mix.Compilers.Elixir do
     end
   end
 
+  @doc """
+  Retrieves all diagnostics from the given manifest.
+  """
+  def diagnostics(manifest, dest) do
+    {_, all_sources, _, _, _, _, _, _} = parse_manifest(manifest, dest)
+    previous_warnings(all_sources, false)
+  end
+
   defp compiler_info_from_force(manifest, all_paths, all_modules, dest) do
     # A config, path dependency or manifest has changed, let's just compile everything
     for {module, _} <- all_modules,
@@ -357,8 +373,7 @@ defmodule Mix.Compilers.Elixir do
             Map.has_key?(stale_modules, module) ->
               {[module | modules_to_recompile], modules_to_mix_check}
 
-            recompile? and Code.ensure_loaded?(module) and
-                function_exported?(module, :__mix_recompile__?, 0) ->
+            recompile? ->
               {modules_to_recompile, [module | modules_to_mix_check]}
 
             true ->
@@ -366,10 +381,18 @@ defmodule Mix.Compilers.Elixir do
           end
       end
 
+    _ = Code.ensure_all_loaded(modules_to_mix_check)
+
     modules_to_recompile =
       modules_to_recompile ++
         for {:ok, {module, true}} <-
-              Task.async_stream(modules_to_mix_check, &{&1, &1.__mix_recompile__?()},
+              Task.async_stream(
+                modules_to_mix_check,
+                fn module ->
+                  {module,
+                   function_exported?(module, :__mix_recompile__?, 0) and
+                     module.__mix_recompile__?()}
+                end,
                 ordered: false,
                 timeout: :infinity
               ) do
@@ -744,21 +767,38 @@ defmodule Mix.Compilers.Elixir do
   end
 
   defp apply_warnings(sources, runtime_warnings, compile_warnings) do
+    cwd = File.cwd!()
     runtime_group = Enum.group_by(runtime_warnings, & &1.source)
     compile_group = Enum.group_by(compile_warnings, & &1.source)
 
     for {source_path, source_entry} <- sources, into: %{} do
-      key = Path.absname(source_path)
-
       source(
         runtime_warnings: runtime_warnings,
-        compile_warnings: compile_warnings
+        compile_warnings: compile_warnings,
+        external: external
       ) = source_entry
 
+      keys = [
+        Path.absname(source_path, cwd)
+        | Enum.map(external, &(&1 |> elem(0) |> Path.absname(cwd)))
+      ]
+
+      runtime_warnings =
+        case Enum.flat_map(keys, &Map.get(runtime_group, &1, [])) do
+          [] -> runtime_warnings
+          runtime_warnings -> runtime_warnings
+        end
+
+      compile_warnings =
+        case Enum.flat_map(keys, &Map.get(compile_group, &1, [])) do
+          [] -> compile_warnings
+          compile_warnings -> compile_warnings
+        end
+
       {source_path,
        source(source_entry,
-         runtime_warnings: Map.get(runtime_group, key, runtime_warnings),
-         compile_warnings: Map.get(compile_group, key, compile_warnings)
+         runtime_warnings: runtime_warnings,
+         compile_warnings: compile_warnings
        )}
     end
   end
@@ -982,6 +1022,24 @@ defmodule Mix.Compilers.Elixir do
     end
   end
 
+  defp modules_diff(compiled_modules, removed_modules, all_modules, timestamp) do
+    {changed, added} =
+      compiled_modules
+      |> Map.keys()
+      |> Enum.split_with(&Map.has_key?(all_modules, &1))
+
+    # Note that removed_modules may also include changed modules
+    removed =
+      for {module, _} <- removed_modules, not Map.has_key?(compiled_modules, module), do: module
+
+    %{
+      added: added,
+      changed: changed,
+      removed: removed,
+      timestamp: timestamp
+    }
+  end
+
   ## Compiler loop
   # The compiler is invoked in a separate process so we avoid blocking its main loop.
 
@@ -991,6 +1049,7 @@ defmodule Mix.Compilers.Elixir do
     threshold = opts[:long_compilation_threshold] || 10
     profile = opts[:profile]
     verbose = opts[:verbose] || false
+    warnings_as_errors = opts[:warnings_as_errors] || false
 
     pid =
       spawn_link(fn ->
@@ -1012,7 +1071,8 @@ defmodule Mix.Compilers.Elixir do
           long_compilation_threshold: threshold,
           profile: profile,
           beam_timestamp: timestamp,
-          return_diagnostics: true
+          return_diagnostics: true,
+          warnings_as_errors: warnings_as_errors
         ]
 
         response = Kernel.ParallelCompiler.compile_to_path(stale, dest, compile_opts)
