@@ -9,7 +9,7 @@ defmodule Mix.SCM.Git do
 
   @impl true
   def format(opts) do
-    if rev = opts[:ref] || opts[:branch] || opts[:tag] do
+    if rev = get_opts_rev(opts) do
       "#{redact_uri(opts[:git])} - #{rev}"
     else
       redact_uri(opts[:git])
@@ -126,17 +126,17 @@ defmodule Mix.SCM.Git do
     update_origin(opts[:git])
 
     # Fetch external data
-    branch_or_tag = opts[:branch] || opts[:tag]
+    rev = get_lock_rev(opts[:lock], opts) || get_opts_rev(opts)
 
     ["--git-dir=.git", "fetch", "--force", "--quiet"]
     |> Kernel.++(progress_switch(git_version()))
     |> Kernel.++(tags_switch(opts[:tag]))
     |> Kernel.++(depth_switch(opts[:depth]))
-    |> Kernel.++(if branch_or_tag, do: ["origin", branch_or_tag], else: [])
+    |> Kernel.++(refspec_switch(opts, rev))
     |> git!()
 
     # Migrate the Git repo
-    rev = get_lock_rev(opts[:lock], opts) || opts[:ref] || branch_or_tag || default_branch()
+    rev = rev || default_branch()
     git!(["--git-dir=.git", "checkout", "--quiet", rev])
 
     if opts[:submodules] do
@@ -197,7 +197,7 @@ defmodule Mix.SCM.Git do
   end
 
   defp ensure_feature_compatibility(version, required_version, feature) do
-    unless required_version <= version do
+    if not (required_version <= version) do
       Mix.raise(
         "Git >= #{format_version(required_version)} is required to use #{feature}. " <>
           "You are running version #{format_version(version)}"
@@ -219,6 +219,15 @@ defmodule Mix.SCM.Git do
     ["--depth=#{n}"]
   end
 
+  defp refspec_switch(_opts, nil), do: []
+
+  defp refspec_switch(opts, rev) do
+    case Keyword.take(opts, [:depth, :branch, :tag]) do
+      [_ | _] -> ["origin", rev]
+      _ -> []
+    end
+  end
+
   ## Helpers
 
   defp validate_git_options(opts) do
@@ -249,25 +258,30 @@ defmodule Mix.SCM.Git do
     end
   end
 
-  defp validate_depth(opts) do
-    case Keyword.take(opts, [:depth, :ref]) do
-      [_, _] ->
-        Mix.raise(
-          "Cannot specify :depth and :ref at the same time. " <>
-            "Error on Git dependency: #{redact_uri(opts[:git])}"
-        )
+  @sha1_size 40
 
-      [depth: depth] when is_integer(depth) and depth > 0 ->
+  defp validate_depth(opts) do
+    case Keyword.take(opts, [:depth]) do
+      [] ->
         opts
 
-      [depth: invalid_depth] ->
+      [{:depth, depth}] when is_integer(depth) and depth > 0 ->
+        ref = opts[:ref]
+
+        if ref && byte_size(ref) < @sha1_size do
+          Mix.raise(
+            "When :depth is used with :ref, a full commit hash is required. " <>
+              "Error on Git dependency: #{redact_uri(opts[:git])}"
+          )
+        end
+
+        opts
+
+      invalid_depth ->
         Mix.raise(
           "The depth must be a positive integer, and be specified only once, got: #{inspect(invalid_depth)}. " <>
             "Error on Git dependency: #{redact_uri(opts[:git])}"
         )
-
-      _ ->
-        opts
     end
   end
 
@@ -296,6 +310,10 @@ defmodule Mix.SCM.Git do
     end
   end
 
+  defp get_opts_rev(opts) do
+    opts[:branch] || opts[:ref] || opts[:tag]
+  end
+
   defp redact_uri(git) do
     case URI.parse(git) do
       %{userinfo: nil} -> git
