@@ -12,6 +12,11 @@ defmodule Mix.Tasks.Deps.Loadpaths do
   Although this task does not show up in `mix help`, it is
   part of Mix public API and can be depended on.
 
+  ## Configuration
+
+    * `:listeners` - the list of listener modules. For more details
+      see `Mix.Task.Compiler`
+
   ## Command line options
 
     * `--no-archives-check` - does not check archives
@@ -19,13 +24,27 @@ defmodule Mix.Tasks.Deps.Loadpaths do
     * `--no-deps-check` - does not check or compile deps, only load available ones
     * `--no-elixir-version-check` - does not check Elixir version
     * `--no-optional-deps` - does not compile or load optional deps
-    * `--no-path-loading` - does not add entries to the code path
 
   """
 
   @impl true
   def run(args) do
-    unless "--no-archives-check" in args do
+    # Note that we need to ensure the dependencies are compiled first,
+    # before we can start the pub/sub listeners, since those come from
+    # the dependencies. Theoretically, between compiling dependencies
+    # and starting the listeners, there may be a concurrent compilation
+    # of the dependencies, which we would miss, and we would already
+    # have modules from our compilation loaded. To avoid this race
+    # condition we start the pub/sub beforehand and we accumulate all
+    # events until the listeners are started. Alternatively we could
+    # use a lock around compilation and sterning the listeners, however
+    # the added benefit of the current approach is that we consistently
+    # receive events for all dependency compilations. Also, if we ever
+    # decide to start the listeners later (e.g. after loadspaths), the
+    # accumulation approach still works.
+    Mix.PubSub.start()
+
+    if "--no-archives-check" not in args do
       Mix.Task.run("archive.check", args)
     end
 
@@ -40,17 +59,19 @@ defmodule Mix.Tasks.Deps.Loadpaths do
 
     config = Mix.Project.config()
 
-    unless "--no-elixir-version-check" in args do
+    if "--no-elixir-version-check" not in args do
       check_elixir_version(config)
     end
 
-    unless "--no-deps-check" in args do
+    if "--no-deps-check" not in args do
       deps_check(all, "--no-compile" in args)
     end
 
-    unless "--no-path-loading" in args do
-      Code.prepend_paths(Enum.flat_map(all, &Mix.Dep.load_paths/1), cache: true)
-    end
+    Code.prepend_paths(Enum.flat_map(all, &Mix.Dep.load_paths/1), cache: true)
+
+    # For now we only allow listeners defined in dependencies, so
+    # we start them right after adding adding deps to the path
+    Mix.PubSub.start_listeners()
 
     :ok
   end
@@ -59,7 +80,7 @@ defmodule Mix.Tasks.Deps.Loadpaths do
     if req = config[:elixir] do
       case Version.parse_requirement(req) do
         {:ok, req} ->
-          unless Version.match?(System.version(), req) do
+          if not Version.match?(System.version(), req) do
             raise Mix.ElixirVersionError,
               target: config[:app] || Mix.Project.get(),
               expected: req,
